@@ -1,5 +1,5 @@
 """
-morecantile is an adpatation of mapbox/mercantile to work with custom projection.
+morecantile is an adaptation of mapbox/mercantile to work with custom projection.
 
 Refs:
     - mapproxy: https://github.com/mapproxy/mapproxy
@@ -8,10 +8,8 @@ Refs:
 
 """
 
-import pkg_resources
+__version__ = "3.1.2"
 
-from .commons import Coords, CoordsBbox, Tile  # noqa
+from .commons import BoundingBox, Coords, Tile  # noqa
 from .defaults import tms  # noqa
 from .models import TileMatrixSet  # noqa
-
-version = pkg_resources.get_distribution(__package__).version
