@@ -1,22 +1,27 @@
 """Pydantic modules for OGC TileMatrixSets (https://www.ogc.org/standards/tms)"""
+
 import math
-import os
 import warnings
 from typing import Any, Dict, Iterator, List, Optional, Sequence, Tuple, Union
 
-from pydantic import AnyHttpUrl, BaseModel, Field, validator
-from rasterio.crs import CRS, epsg_treats_as_latlong, epsg_treats_as_northingeasting
-from rasterio.features import bounds as feature_bounds
-from rasterio.warp import transform, transform_bounds, transform_geom
+from pydantic import AnyHttpUrl, BaseModel, Field, PrivateAttr, validator
+from pyproj import CRS, Transformer
+from pyproj.enums import WktVersion
+from pyproj.exceptions import ProjError
 
-from .commons import Coords, CoordsBbox, Tile
-from .errors import DeprecationWarning, InvalidIdentifier
+from .commons import BoundingBox, Coords, Tile
+from .errors import (
+    InvalidZoomError,
+    NoQuadkeySupport,
+    PointOutsideTMSBounds,
+    QuadKeyError,
+)
 from .utils import (
     _parse_tile_arg,
     bbox_to_feature,
+    check_quadkey_support,
     meters_per_unit,
     point_in_bbox,
-    truncate_lnglat,
 )
 
 NumType = Union[float, int]
@@ -25,6 +30,45 @@ LL_EPSILON = 1e-11
 WGS84_CRS = CRS.from_epsg(4326)
 
 
+class CRSType(CRS, str):
+    """
+    A geographic or projected coordinate reference system.
+    """
+
+    @classmethod
+    def __get_validators__(cls):
+        """validator for the type."""
+        yield cls.validate
+
+    @classmethod
+    def validate(cls, value: Union[CRS, str]) -> CRS:
+        """Validate CRS."""
+        # If input is a string we tranlate it to CRS
+        if not isinstance(value, CRS):
+            return CRS.from_user_input(value)
+
+        return value
+
+    @classmethod
+    def __modify_schema__(cls, field_schema):
+        """Update default schema."""
+        field_schema.update(
+            anyOf=[
+                {"type": "pyproj.CRS"},
+                {"type": "string", "minLength": 1, "maxLength": 65536},
+            ],
+            examples=[
+                "CRS.from_epsg(4326)",
+                "http://www.opengis.net/def/crs/EPSG/0/3978",
+                "urn:ogc:def:crs:EPSG::2193",
+            ],
+        )
+
+    def __repr__(self):
+        """Type representation."""
+        return f"CRS({super().__repr__()})"
+
+
 def CRS_to_uri(crs: CRS) -> str:
     """Convert CRS to URI."""
     epsg_code = crs.to_epsg()
@@ -33,30 +77,23 @@ def CRS_to_uri(crs: CRS) -> str:
 
 def crs_axis_inverted(crs: CRS) -> bool:
     """Check if CRS has inverted AXIS (lat,lon) instead of (lon,lat)."""
-    return epsg_treats_as_latlong(crs) or epsg_treats_as_northingeasting(crs)
+    return crs.axis_info[0].abbrev.upper() in ["Y", "LAT", "N"]
 
 
-class BoundingBox(BaseModel):
+class TMSBoundingBox(BaseModel):
     """Bounding box"""
 
     type: str = Field("BoundingBoxType", const=True)
-    crs: Union[CRS, AnyHttpUrl]
+    crs: CRSType
     lowerCorner: BoundsType
     upperCorner: BoundsType
 
     class Config:
-        """Configure BoundingBox."""
+        """Configure TMSBoundingBox."""
 
         arbitrary_types_allowed = True
         json_encoders = {CRS: lambda v: CRS_to_uri(v)}
 
-    @validator("crs")
-    def validate_crs(cls, crs):
-        """Translate URI to rasterio CRS Object."""
-        if not isinstance(crs, CRS):
-            crs = CRS.from_user_input(crs)
-        return crs
-
 
 class TileMatrix(BaseModel):
     """Tile matrix"""
@@ -87,39 +124,80 @@ class TileMatrixSet(BaseModel):
     abstract: Optional[str]
     keywords: Optional[List[str]]
     identifier: str = Field(..., regex=r"^[\w\d_\-]+$")
-    supportedCRS: Union[CRS, AnyHttpUrl]
+    supportedCRS: CRSType
     wellKnownScaleSet: Optional[AnyHttpUrl] = None
-    boundingBox: Optional[BoundingBox]
+    boundingBox: Optional[TMSBoundingBox]
     tileMatrix: List[TileMatrix]
 
+    # Private attributes
+    _is_quadtree: bool = PrivateAttr()
+
+    # CRS transformation attributes
+    _geographic_crs: CRSType = PrivateAttr(default=WGS84_CRS)
+    _to_geographic: Transformer = PrivateAttr()
+    _from_geographic: Transformer = PrivateAttr()
+
     class Config:
         """Configure TileMatrixSet."""
 
         arbitrary_types_allowed = True
         json_encoders = {CRS: lambda v: CRS_to_uri(v)}
 
+    def __init__(self, **data):
+        """Create PyProj transforms and check if TileMatrixSet supports quadkeys."""
+        super().__init__(**data)
+
+        self._is_quadtree = check_quadkey_support(self.tileMatrix)
+
+        self._geographic_crs = data.get("_geographic_crs", WGS84_CRS)
+
+        try:
+            self._to_geographic = Transformer.from_crs(
+                self.supportedCRS, self._geographic_crs, always_xy=True
+            )
+            self._from_geographic = Transformer.from_crs(
+                self._geographic_crs, self.supportedCRS, always_xy=True
+            )
+        except ProjError:
+            warnings.warn(
+                "Could not create coordinate Transformer from input CRS to the given geographic CRS"
+                "some methods might not be available.",
+                UserWarning,
+            )
+            self._to_geographic = None
+            self._from_geographic = None
+
     @validator("tileMatrix")
     def sort_tile_matrices(cls, v):
         """Sort matrices by identifier"""
         return sorted(v, key=lambda m: int(m.identifier))
 
-    @validator("supportedCRS")
-    def validate_crs(cls, crs):
-        """Translate URI to rasterio CRS Object."""
-        if not isinstance(crs, CRS):
-            crs = CRS.from_user_input(crs)
-        return crs
-
     def __iter__(self):
         """Iterate over matrices"""
         for matrix in self.tileMatrix:
             yield matrix
 
+    def __repr__(self):
+        """Simplify default pydantic model repr."""
+        return f"<TileMatrixSet title='{self.title}' identifier='{self.identifier}'>"
+
     @property
     def crs(self) -> CRS:
         """Fetch CRS from epsg"""
         return self.supportedCRS
 
+    @property
+    def rasterio_crs(self):
+        """Return rasterio CRS."""
+
+        import rasterio
+        from rasterio.env import GDALVersion
+
+        if GDALVersion.runtime().major < 3:
+            return rasterio.crs.CRS.from_wkt(self.crs.to_wkt(WktVersion.WKT1_GDAL))
+        else:
+            return rasterio.crs.CRS.from_wkt(self.crs.to_wkt())
+
     @property
     def minzoom(self) -> int:
         """TileMatrixSet minimum TileMatrix identifier"""
@@ -135,18 +213,6 @@ class TileMatrixSet(BaseModel):
         """Check if CRS has inverted AXIS (lat,lon) instead of (lon,lat)."""
         return crs_axis_inverted(self.crs)
 
-    @classmethod
-    def load(cls, name: str):
-        """Load default TileMatrixSet."""
-        warnings.warn(
-            "TileMatrixSet.load will be deprecated in version 2.0.0", DeprecationWarning
-        )
-        try:
-            data_dir = os.path.join(os.path.dirname(__file__), "data")
-            return cls.parse_file(os.path.join(data_dir, f"{name}.json"))
-        except FileNotFoundError:
-            raise InvalidIdentifier(f"'{name}' is not a valid default TileMatrixSet.")
-
     @classmethod
     def custom(
         cls,
@@ -160,13 +226,14 @@ class TileMatrixSet(BaseModel):
         maxzoom: int = 24,
         title: str = "Custom TileMatrixSet",
         identifier: str = "Custom",
+        geographic_crs: CRS = WGS84_CRS,
     ):
         """
         Construct a custom TileMatrixSet.
 
         Attributes
         ----------
-        crs: rasterio.crs.CRS
+        crs: pyproj.CRS
             Tile Matrix Set coordinate reference system
         extent: list
             Bounding box of the Tile Matrix Set, (left, bottom, right, top).
@@ -178,8 +245,8 @@ class TileMatrixSet(BaseModel):
             Tiling schema coalescence coefficient (default: [1, 1] for EPSG:3857).
             Should be set to [2, 1] for EPSG:4326.
             see: http://docs.opengeospatial.org/is/17-083r2/17-083r2.html#14
-        extent_crs: rasterio.crs.CRS
-            Extent's coordinate reference system, as a rasterio CRS object.
+        extent_crs: pyproj.CRS
+            Extent's coordinate reference system, as a pyproj CRS object.
             (default: same as input crs)
         minzoom: int
             Tile Matrix Set minimum zoom level (default is 0).
@@ -189,6 +256,8 @@ class TileMatrixSet(BaseModel):
             Tile Matrix Set title (default is 'Custom TileMatrixSet')
         identifier: str
             Tile Matrix Set identifier (default is 'Custom')
+        geographic_crs: pyproj.CRS
+            Geographic (lat,lon) coordinate reference system (default is EPSG:4326)
 
         Returns:
         --------
@@ -200,38 +269,38 @@ class TileMatrixSet(BaseModel):
             "identifier": identifier,
             "supportedCRS": crs,
             "tileMatrix": [],
+            "_geographic_crs": geographic_crs,
         }
 
-        is_inverted = False
-        if crs.to_epsg():
-            # We use URI because with EPSG code it doesn't work in GDAL 2
-            crs_uri = CRS_to_uri(crs)
-            is_inverted = crs_axis_inverted(CRS.from_user_input(crs_uri))
+        is_inverted = crs_axis_inverted(crs)
 
-        tms["boundingBox"] = BoundingBox(
-            **dict(
+        if is_inverted:
+            tms["boundingBox"] = TMSBoundingBox(
                 crs=extent_crs or crs,
-                lowerCorner=[extent[0], extent[1]]
-                if not is_inverted
-                else [extent[1], extent[0]],
-                upperCorner=[extent[2], extent[3]]
-                if not is_inverted
-                else [extent[2], extent[3]],
+                lowerCorner=[extent[1], extent[0]],
+                upperCorner=[extent[3], extent[2]],
             )
-        )
-
-        bbox = CoordsBbox(
-            *(
-                transform_bounds(extent_crs, crs, *extent, densify_pts=21)
-                if extent_crs
-                else extent
+        else:
+            tms["boundingBox"] = TMSBoundingBox(
+                crs=extent_crs or crs,
+                lowerCorner=[extent[0], extent[1]],
+                upperCorner=[extent[2], extent[3]],
             )
-        )
-        x_origin = bbox.xmin if not is_inverted else bbox.ymax
-        y_origin = bbox.ymax if not is_inverted else bbox.xmin
 
-        width = abs(bbox.xmax - bbox.xmin)
-        height = abs(bbox.ymax - bbox.ymin)
+        if extent_crs:
+            transform = Transformer.from_crs(extent_crs, crs, always_xy=True)
+            left, bottom, right, top = extent
+            bbox = BoundingBox(
+                *transform.transform_bounds(left, bottom, right, top, densify_pts=21)
+            )
+        else:
+            bbox = BoundingBox(*extent)
+
+        x_origin = bbox.left if not is_inverted else bbox.top
+        y_origin = bbox.top if not is_inverted else bbox.left
+
+        width = abs(bbox.right - bbox.left)
+        height = abs(bbox.top - bbox.bottom)
         mpu = meters_per_unit(crs)
         for zoom in range(minzoom, maxzoom + 1):
             res = max(
@@ -309,12 +378,24 @@ class TileMatrixSet(BaseModel):
         """
         return matrix.scaleDenominator * 0.28e-3 / meters_per_unit(self.crs)
 
-    def zoom_for_res(self, res: float, max_z: Optional[int] = None) -> int:
+    def zoom_for_res(
+        self,
+        res: float,
+        max_z: Optional[int] = None,
+        zoom_level_strategy: str = "auto",
+        min_z: Optional[int] = None,
+    ) -> int:
         """Get TMS zoom level corresponding to a specific resolution.
 
         Args:
             res (float): Resolution in TMS unit.
             max_z (int): Maximum zoom level (default is tms maxzoom).
+            zoom_level_strategy (str): Strategy to determine zoom level (same as in GDAL 3.2).
+                LOWER will select the zoom level immediately below the theoretical computed non-integral zoom level.
+                On the contrary, UPPER will select the immediately above zoom level.
+                Defaults to AUTO which selects the closest zoom level.
+                ref: https://gdal.org/drivers/raster/cog.html#raster-cog
+            min_z (int): Minimum zoom level (default is tms minzoom).
 
         Returns:
             int: TMS zoom for a given resolution.
@@ -323,54 +404,94 @@ class TileMatrixSet(BaseModel):
             >>> zoom_for_res(430.021)
 
         """
-        if not max_z:
+        if max_z is None:
             max_z = self.maxzoom
 
-        for z in range(max_z):
-            matrix = self.matrix(z)
-            if res > self._resolution(matrix):
-                return max(0, z - 1)  # We don't want to scale up
+        if min_z is None:
+            min_z = self.minzoom
 
-        return max_z
+        # Freely adapted from https://github.com/OSGeo/gdal/blob/dc38aa64d779ecc45e3cd15b1817b83216cf96b8/gdal/frmts/gtiff/cogdriver.cpp#L272-L305
+        for zoom_level in range(min_z, max_z + 1):
+            matrix_res = self._resolution(self.matrix(zoom_level))
+            if res > matrix_res or abs(res - matrix_res) / matrix_res <= 1e-8:
+                break
+
+        if zoom_level > 0 and abs(res - matrix_res) / matrix_res > 1e-8:
+            if zoom_level_strategy.lower() == "lower":
+                zoom_level = max(zoom_level - 1, min_z)
+            elif zoom_level_strategy.lower() == "upper":
+                zoom_level = min(zoom_level, max_z)
+            elif zoom_level_strategy.lower() == "auto":
+                if (self._resolution(self.matrix(max(zoom_level - 1, min_z))) / res) < (
+                    res / matrix_res
+                ):
+                    zoom_level = max(zoom_level - 1, min_z)
+            else:
+                raise ValueError(
+                    f"Invalid strategy: {zoom_level_strategy}. Should be one of lower|upper|auto"
+                )
+
+        return zoom_level
 
     def lnglat(self, x: float, y: float, truncate=False) -> Coords:
-        """Transform point(x,y) to longitude and latitude."""
-        inside = point_in_bbox(Coords(x, y), self.bbox)
+        """Transform point(x,y) to geographic longitude and latitude."""
+        inside = point_in_bbox(Coords(x, y), self.xy_bbox)
         if not inside:
-            warnings.warn("Point is outside TMS bounds.", UserWarning)
+            warnings.warn(
+                f"Point ({x}, {y}) is outside TMS bounds {list(self.xy_bbox)}.",
+                PointOutsideTMSBounds,
+            )
 
-        xs, ys = transform(self.crs, WGS84_CRS, [x], [y])
-        lng, lat = xs[0], ys[0]
+        lng, lat = self._to_geographic.transform(x, y)
 
         if truncate:
-            lng, lat = truncate_lnglat(lng, lat)
+            lng, lat = self.truncate_lnglat(lng, lat)
 
         return Coords(lng, lat)
 
     def xy(self, lng: float, lat: float, truncate=False) -> Coords:
-        """Transform longitude and latitude coordinates to TMS CRS."""
+        """Transform geographic longitude and latitude coordinates to TMS CRS."""
         if truncate:
-            lng, lat = truncate_lnglat(lng, lat)
+            lng, lat = self.truncate_lnglat(lng, lat)
 
-        xs, ys = transform(WGS84_CRS, self.crs, [lng], [lat])
-        x, y = xs[0], ys[0]
+        inside = point_in_bbox(Coords(lng, lat), self.bbox)
+        if not inside:
+            warnings.warn(
+                f"Point ({lng}, {lat}) is outside TMS bounds {list(self.bbox)}.",
+                PointOutsideTMSBounds,
+            )
 
-        # https://github.com/mapbox/mercantile/blob/master/mercantile/__init__.py#L232-L237
-        if lat <= -90:
-            y = float("-inf")
-        elif lat >= 90:
-            y = float("inf")
+        x, y = self._from_geographic.transform(lng, lat)
 
         return Coords(x, y)
 
+    def truncate_lnglat(self, lng: float, lat: float) -> Tuple[float, float]:
+        """
+        Truncate geographic coordinates to TMS geographic bbox.
+
+        Adapted from https://github.com/mapbox/mercantile/blob/master/mercantile/__init__.py
+
+        """
+        if lng > self.bbox.right:
+            lng = self.bbox.right
+        elif lng < self.bbox.left:
+            lng = self.bbox.left
+
+        if lat > self.bbox.top:
+            lat = self.bbox.top
+        elif lat < self.bbox.bottom:
+            lat = self.bbox.bottom
+
+        return lng, lat
+
     def _tile(self, xcoord: float, ycoord: float, zoom: int) -> Tile:
         """
-        Get the tile containing a Point (in input projection).
+        Get the tile containing a Point (in TMS CRS).
 
         Parameters
         ----------
         xcoord, ycoord : float
-            A longitude and latitude pair in input projection.
+            A `X` and `Y` pair in TMS coordinate reference system.
         zoom : int
             The zoom level.
 
@@ -417,16 +538,16 @@ class TileMatrixSet(BaseModel):
 
     def tile(self, lng: float, lat: float, zoom: int, truncate=False) -> Tile:
         """
-        Get the tile containing a longitude and latitude.
+        Get the tile for a given geographic longitude and latitude pair.
 
         Parameters
         ----------
         lng, lat : float
-            A longitude and latitude pair in decimal degrees.
+            A longitude and latitude pair in geographic coordinate reference system.
         zoom : int
-            The web mercator zoom level.
+            The zoom level.
         truncate : bool
-            Whether or not to truncate inputs to limits of WGS84 bounds.
+            Whether or not to truncate inputs to limits of TMS geographic bounds.
 
         Returns
         -------
@@ -438,19 +559,20 @@ class TileMatrixSet(BaseModel):
 
     def _ul(self, *tile: Tile) -> Coords:
         """
-        Return the upper left coordinate of the (x, y, z) tile in input projection.
+        Return the upper left coordinate of the tile in TMS coordinate reference system.
 
         Attributes
         ----------
-        tile: (x, y, z) tile coordinates or a Tile object we want the upper left geospatial coordinates of.
+        tile: (x, y, z) tile coordinates or a Tile object we want the upper left coordinates of.
 
         Returns
         -------
-        The upper left geospatial coordiantes of the input tile.
+        Coords: The upper left coordinates of the input tile.
 
         """
-        tile = _parse_tile_arg(*tile)
-        matrix = self.matrix(tile.z)
+        t = _parse_tile_arg(*tile)
+
+        matrix = self.matrix(t.z)
         res = self._resolution(matrix)
 
         origin_x = (
@@ -460,13 +582,13 @@ class TileMatrixSet(BaseModel):
             matrix.topLeftCorner[0] if self._invert_axis else matrix.topLeftCorner[1]
         )
 
-        xcoord = origin_x + tile.x * res * matrix.tileWidth
-        ycoord = origin_y - tile.y * res * matrix.tileHeight
+        xcoord = origin_x + t.x * res * matrix.tileWidth
+        ycoord = origin_y - t.y * res * matrix.tileHeight
         return Coords(xcoord, ycoord)
 
-    def xy_bounds(self, *tile: Tile) -> CoordsBbox:
+    def xy_bounds(self, *tile: Tile) -> BoundingBox:
         """
-        Return the bounding box of the (x, y, z) tile in input projection.
+        Return the bounding box of the tile in TMS coordinate reference system.
 
         Attributes
         ----------
@@ -474,47 +596,51 @@ class TileMatrixSet(BaseModel):
 
         Returns
         -------
-        The bounding box of the input tile.
+        BoundingBox: The bounding box of the input tile.
 
         """
-        tile = _parse_tile_arg(*tile)
-        left, top = self._ul(*tile)
-        right, bottom = self._ul(tile.x + 1, tile.y + 1, tile.z)
-        return CoordsBbox(left, bottom, right, top)
+        t = _parse_tile_arg(*tile)
+
+        left, top = self._ul(t)
+        right, bottom = self._ul(Tile(t.x + 1, t.y + 1, t.z))
+        return BoundingBox(left, bottom, right, top)
 
     def ul(self, *tile: Tile) -> Coords:
         """
-        Return the upper left coordinate of the (x, y, z) tile in Lat Lon.
+        Return the upper left coordinates of the tile in geographic coordinate reference system.
 
         Attributes
         ----------
-        tile: (x, y, z) tile coordinates or a Tile object we want the upper left geospatial coordinates of.
+        tile (tuple or Tile): (x, y, z) tile coordinates or a Tile object we want the upper left geographic coordinates of.
 
         Returns
         -------
-        The upper left geospatial coordiantes of the input tile.
+        Coords: The upper left geographic coordinates of the input tile.
 
         """
-        x, y = self._ul(*tile)
+        t = _parse_tile_arg(*tile)
+
+        x, y = self._ul(t)
         return Coords(*self.lnglat(x, y))
 
-    def bounds(self, *tile: Tile) -> CoordsBbox:
+    def bounds(self, *tile: Tile) -> BoundingBox:
         """
-        Return the bounding box of the (x, y, z) tile in LatLong.
+        Return the bounding box of the tile in geographic coordinate reference system.
 
         Attributes
         ----------
-        tile: A tuple of (x, y, z) tile coordinates or a Tile object we want the bounding box of.
+        tile (tuple or Tile): A tuple of (x, y, z) tile coordinates or a Tile object we want the bounding box of.
 
         Returns
         -------
-        The bounding box of the input tile.
+        BoundingBox: The bounding box of the input tile.
 
         """
-        tile = _parse_tile_arg(*tile)
-        left, top = self.ul(tile.x, tile.y, tile.z)
-        right, bottom = self.ul(tile.x + 1, tile.y + 1, tile.z)
-        return CoordsBbox(left, bottom, right, top)
+        t = _parse_tile_arg(*tile)
+
+        left, top = self.ul(t)
+        right, bottom = self.ul(Tile(t.x + 1, t.y + 1, t.z))
+        return BoundingBox(left, bottom, right, top)
 
     @property
     def xy_bbox(self):
@@ -540,23 +666,43 @@ class TileMatrixSet(BaseModel):
                 if self._invert_axis
                 else self.boundingBox.upperCorner[1]
             )
+            if self.boundingBox.crs != self.crs:
+                transform = Transformer.from_crs(
+                    self.boundingBox.crs, self.crs, always_xy=True
+                )
+                left, bottom, right, top = transform.transform_bounds(
+                    left,
+                    bottom,
+                    right,
+                    top,
+                    densify_pts=21,
+                )
+
         else:
             zoom = self.minzoom
             matrix = self.matrix(zoom)
-            left, top = self._ul(0, 0, zoom)
-            right, bottom = self._ul(matrix.matrixWidth, matrix.matrixHeight, zoom)
+            left, top = self._ul(Tile(0, 0, zoom))
+            right, bottom = self._ul(
+                Tile(matrix.matrixWidth, matrix.matrixHeight, zoom)
+            )
 
-        return CoordsBbox(left, bottom, right, top)
+        return BoundingBox(left, bottom, right, top)
 
     @property
     def bbox(self):
-        """Return TMS bounding box in WGS84."""
-        left, bottom, right, top = transform_bounds(
-            self.crs, WGS84_CRS, *self.xy_bbox, densify_pts=21
+        """Return TMS bounding box in geographic coordinate reference system."""
+        left, bottom, right, top = self.xy_bbox
+        return BoundingBox(
+            *self._to_geographic.transform_bounds(
+                left,
+                bottom,
+                right,
+                top,
+                densify_pts=21,
+            )
         )
-        return CoordsBbox(left, bottom, right, top)
 
-    def intersect_tms(self, bbox: CoordsBbox) -> bool:
+    def intersect_tms(self, bbox: BoundingBox) -> bool:
         """Check if a bounds intersects with the TMS bounds."""
         tms_bounds = self.xy_bbox
         return (
@@ -583,7 +729,7 @@ class TileMatrixSet(BaseModel):
         Parameters
         ----------
         west, south, east, north : sequence of float
-            Bounding values in decimal degrees.
+            Bounding values in decimal degrees (geographic CRS).
         zooms : int or sequence of int
             One or more zoom levels.
         truncate : bool, optional
@@ -603,28 +749,27 @@ class TileMatrixSet(BaseModel):
             zooms = (zooms,)
 
         if truncate:
-            west, south = truncate_lnglat(west, south)
-            east, north = truncate_lnglat(east, north)
+            west, south = self.truncate_lnglat(west, south)
+            east, north = self.truncate_lnglat(east, north)
 
         if west > east:
-            bbox_west = (-180.0, south, east, north)
-            bbox_east = (west, south, 180.0, north)
+            bbox_west = (self.bbox.left, south, east, north)
+            bbox_east = (west, south, self.bbox.right, north)
             bboxes = [bbox_west, bbox_east]
         else:
             bboxes = [(west, south, east, north)]
 
-        # tms_bounds = self._tms_bounds
         for w, s, e, n in bboxes:
-            # w, s, e, n = transform_bounds(WGS84_CRS, self.crs, w, s, e, n, densify_pts=21)
-            # if not self.intersect_tms(w, s, e, n):
-            #     continue
+            # Clamp bounding values.
+            w = max(self.bbox.left, w)
+            s = max(self.bbox.bottom, s)
+            e = min(self.bbox.right, e)
+            n = min(self.bbox.top, n)
 
-            # w = max(tms_bounds[0], w)
-            # s = max(tms_bounds[1], s)
-            # e = min(tms_bounds[2], e)
-            # n = min(tms_bounds[3], n)
             for z in zooms:
-                ul_tile = self.tile(w, n, z)
+                ul_tile = self.tile(
+                    w + LL_EPSILON, n - LL_EPSILON, z
+                )  # Not in mercantile
                 lr_tile = self.tile(e - LL_EPSILON, s + LL_EPSILON, z)
 
                 for i in range(ul_tile.x, lr_tile.x + 1):
@@ -643,7 +788,7 @@ class TileMatrixSet(BaseModel):
         """
         Get the GeoJSON feature corresponding to a tile.
 
-        Originaly from https://github.com/mapbox/mercantile/blob/master/mercantile/__init__.py
+        Originally from https://github.com/mapbox/mercantile/blob/master/mercantile/__init__.py
 
         Parameters
         ----------
@@ -669,9 +814,9 @@ class TileMatrixSet(BaseModel):
         west, south, east, north = self.xy_bounds(tile)
 
         if not projected:
-            geom = bbox_to_feature(west, south, east, north)
-            geom = transform_geom(self.crs, WGS84_CRS, geom)
-            west, south, east, north = feature_bounds(geom)
+            west, south, east, north = self._to_geographic.transform_bounds(
+                west, south, east, north, densify_pts=21
+            )
 
         if buffer:
             west -= buffer
@@ -717,3 +862,220 @@ class TileMatrixSet(BaseModel):
             feat["id"] = fid
 
         return feat
+
+    def quadkey(self, *tile: Tile) -> str:
+        """Get the quadkey of a tile
+
+        Parameters
+        ----------
+        tile : Tile or sequence of int
+            May be be either an instance of Tile or 3 ints, X, Y, Z.
+
+        Returns
+        -------
+        str
+
+        """
+        if not self._is_quadtree:
+            raise NoQuadkeySupport(
+                "This Tile Matrix Set doesn't support 2 x 2 quadkeys."
+            )
+
+        t = _parse_tile_arg(*tile)
+
+        qk = []
+        for z in range(t.z, self.minzoom, -1):
+            digit = 0
+            mask = 1 << (z - 1)
+            if t.x & mask:
+                digit += 1
+            if t.y & mask:
+                digit += 2
+            qk.append(str(digit))
+
+        return "".join(qk)
+
+    def quadkey_to_tile(self, qk: str) -> Tile:
+        """Get the tile corresponding to a quadkey
+
+        Parameters
+        ----------
+        qk : str
+            A quadkey string.
+
+        Returns
+        -------
+        Tile
+
+        """
+        if not self._is_quadtree:
+            raise NoQuadkeySupport(
+                "This Tile Matrix Set doesn't support 2 x 2 quadkeys."
+            )
+
+        if len(qk) == 0:
+            return Tile(0, 0, 0)
+
+        xtile, ytile = 0, 0
+        for i, digit in enumerate(reversed(qk)):
+            mask = 1 << i
+            if digit == "1":
+                xtile = xtile | mask
+            elif digit == "2":
+                ytile = ytile | mask
+            elif digit == "3":
+                xtile = xtile | mask
+                ytile = ytile | mask
+            elif digit != "0":
+                raise QuadKeyError("Unexpected quadkey digit: %r", digit)
+
+        return Tile(xtile, ytile, i + 1)
+
+    def minmax(self, zoom: int) -> Dict:
+        """Return TileMatrix Extrema.
+
+        Parameters
+        ----------
+        zoom : int
+            The zoom level.
+
+        Returns
+        -------
+        Dict
+
+        """
+        m = self.matrix(zoom)
+        return {
+            "x": {"min": 0, "max": m.matrixWidth - 1},
+            "y": {"min": 0, "max": m.matrixHeight - 1},
+        }
+
+    def is_valid(self, *tile: Tile) -> bool:
+        """Check if a tile is valid."""
+        t = _parse_tile_arg(*tile)
+
+        if t.z < self.minzoom:
+            return False
+
+        extrema = self.minmax(t.z)
+        validx = extrema["x"]["min"] <= t.x <= extrema["x"]["max"]
+        validy = extrema["y"]["min"] <= t.y <= extrema["y"]["max"]
+
+        return validx and validy
+
+    def neighbors(self, *tile: Tile) -> List[Tile]:
+        """The neighbors of a tile
+
+        The neighbors function makes no guarantees regarding neighbor tile
+        ordering.
+
+        The neighbors function returns up to eight neighboring tiles, where
+        tiles will be omitted when they are not valid.
+
+        Parameters
+        ----------
+        tile : Tile or sequence of int
+            May be be either an instance of Tile or 3 ints, X, Y, Z.
+
+        Returns
+        -------
+        list
+
+        """
+        t = _parse_tile_arg(*tile)
+        extrema = self.minmax(t.z)
+
+        tiles = []
+        for i in [-1, 0, 1]:
+            for j in [-1, 0, 1]:
+                if i == 0 and j == 0:
+                    continue
+                elif t.x + i < extrema["x"]["min"] or t.y + j < extrema["y"]["min"]:
+                    continue
+
+                elif t.x + i > extrema["x"]["max"] or t.y + j > extrema["y"]["max"]:
+                    continue
+
+                tiles.append(Tile(x=t.x + i, y=t.y + j, z=t.z))
+
+        return tiles
+
+    def parent(self, *tile: Tile, zoom: int = None):
+        """Get the parent of a tile
+
+        The parent is the tile of one zoom level lower that contains the
+        given "child" tile.
+
+        Parameters
+        ----------
+        tile : Tile or sequence of int
+            May be be either an instance of Tile or 3 ints, X, Y, Z.
+        zoom : int, optional
+            Determines the *zoom* level of the returned parent tile.
+            This defaults to one lower than the tile (the immediate parent).
+
+        Returns
+        -------
+        list: list of Tile
+
+        """
+        t = _parse_tile_arg(*tile)
+
+        if t.z == self.minzoom:
+            return []
+
+        if zoom is not None and t.z <= zoom:
+            raise InvalidZoomError("zoom must be less than that of the input tile")
+
+        target_zoom = t.z - 1 if zoom is None else zoom
+
+        res = self._resolution(self.matrix(t.z)) / 10.0
+
+        bbox = self.xy_bounds(t)
+        ul_tile = self._tile(bbox.left + res, bbox.top - res, target_zoom)
+        lr_tile = self._tile(bbox.right - res, bbox.bottom + res, target_zoom)
+
+        tiles = []
+        for i in range(ul_tile.x, lr_tile.x + 1):
+            for j in range(ul_tile.y, lr_tile.y + 1):
+                tiles.append(Tile(i, j, target_zoom))
+
+        return tiles
+
+    def children(self, *tile: Tile, zoom: int = None):
+        """Get the children of a tile
+
+        The children are ordered: top-left, top-right, bottom-right, bottom-left.
+
+        Parameters
+        ----------
+        tile : Tile or sequence of int
+            May be be either an instance of Tile or 3 ints, X, Y, Z.
+        zoom : int, optional
+            Determines the *zoom* level of the returned parent tile.
+            This defaults to one lower than the tile (the immediate parent).
+
+        Returns
+        -------
+        list: list of Tile
+
+        """
+        t = _parse_tile_arg(*tile)
+
+        if zoom is not None and t.z > zoom:
+            raise InvalidZoomError("zoom must be greater than that of the input tile")
+
+        target_zoom = t.z + 1 if zoom is None else zoom
+
+        bbox = self.xy_bounds(t)
+        res = self._resolution(self.matrix(t.z)) / 10.0
+
+        ul_tile = self._tile(bbox.left + res, bbox.top - res, target_zoom)
+        lr_tile = self._tile(bbox.right - res, bbox.bottom + res, target_zoom)
+
+        tiles = []
+        for i in range(ul_tile.x, lr_tile.x + 1):
+            for j in range(ul_tile.y, lr_tile.y + 1):
+                tiles.append(Tile(i, j, target_zoom))
+
+        return tiles
